/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *
 * cones.c
 *
 * routines for cones in Geomview
 *
 * $Id: cones.c,v 1.2 2003/01/02 01:43:31 lloyd Exp $
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "globals.h"
#include "constants.h"
#include "gv_utils.h"
#include "sats.h"
#include "orbit_utils.h"
#include "savi.h"

static unsigned int cones_on_flag = FALSE;
static unsigned int cones_geom_exists = FALSE;
static unsigned int cones_geom_needs_updating = FALSE;
static int cones_angle_used = 0;

static void
cone_size(double *radius, double *height, const Satellite s, const
	  Constellation * pconstellation);
static void
cone_write_geom(const Satellite s, const Constellation * pconstellation);

/*
 * cones_on_cmd()
 *
 * Routine to display cones
 */
char *
cones_on_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (cones_on_flag)
    return EMPTY_str;
  cones_on_flag = TRUE;
  cones_geom_needs_updating = TRUE;

  transforms_needed |= (1 << CONES);
  gv_trans_list_create(sl);
  /* and send them too? */

  gv_begin();
  while (sl) {
    cone_display(sl->s, get_constellation());
    sl = sl->next;
  }
  gv_end();

  /* remember what angle was used */
  cones_angle_used = coverage_angle;

  return EMPTY_str;
}

/*
 * cones_off_cmd()
 *
 * Routine to turn off display of cones
 */
char *
cones_off_cmd(int argc, char *argv[])
{
  Satellite_list sl = get_constellation()->satellites;

  if (!cones_on_flag)
    return EMPTY_str;

  transforms_needed &= ~(1 << CONES);

  gv_begin();
  cones_gv_delete();
  while (sl) {
    cone_delete(sl->s);
    sl = sl->next;
  }
  cones_geom_needs_updating = FALSE;
  cones_on_flag = FALSE;
  gv_end();

  return EMPTY_str;
}


char *
cones_set_cmd(int argc, char *argv[])
{
  Constellation *pconstellation = get_constellation();
  Satellite s;

  int n = atoi(argv[2]);
  unsigned int flag = atoi(argv[3]);

  s = constellation_search(pconstellation, n);
  if (s) {
    if (flag)
      s->can_display_cone = 1;
    else
      s->can_display_cone = 0;
    cones_geom_needs_updating = TRUE;
  }

  return EMPTY_str;
}


/*
 * cones_rebuild()
 *
 * Routine to resize all the cones currently built
 */
void
cones_rebuild()
{
  Satellite_list sl = get_constellation()->satellites;

  if (!cones_on_flag)
    return;

  gv_begin();
  while (sl) {
    cone_display(sl->s, get_constellation());
    sl = sl->next;
  }
  gv_end();

  /* remember what angle was used */
  cones_angle_used = coverage_angle;

  cones_geom_needs_updating = FALSE;
}


/*
 * cones_gv_delete
 *
 * routine to delete geomview cones object
 */
void
cones_gv_delete()
{
  if (cones_geom_exists) {
    gv_begin();
    gv_delete_geom("Cones");
    gv_end();
    cones_geom_exists = FALSE;
  }
}


/*
 * cones_relocate
 *
 * Called to update the cone when the satellites have all been moved
 *
 */
void
cones_relocate(const Constellation * pconstellation)
{
  Satellite_list sl = pconstellation->satellites;
  unsigned int angle_changed = (cones_angle_used != coverage_angle);

  if (!cones_on_flag)
    return;

  while (sl) {
    if (angle_changed || (sl->s->oe.e > 0)) {
      cone_write_geom(sl->s, pconstellation);
    }
    sl = sl->next;
  }

  /* remember what angle was used */
  cones_angle_used = coverage_angle;
}

/*
 * cone_size
 *
 * computes radius and height of cone for given satellite
 *
 */
static void
cone_size(double *radius, double *height, const Satellite s, const
	  Constellation * pconstellation)
{
  CentralBody *pcb = pconstellation->pcb;
  double r, theta, c, si, temp;

  r = s->x_S.r / pcb->radius;
  if (r <= 1.0) {
    /* satellite inside central body! */
    *radius = 0;
    *height = 0;
    return;
  }

  /* coverage_angle is global variable (in degrees) */
  theta = coverage_angle * DEG_TO_RAD;
  c = cos(theta);
  si = sin(theta);
  if (MASK_ELEVATION == get_coverage_type()) {
    temp = sqrt(r * r - c * c);
    *radius = c * (temp - si) / r;
    *height = temp * (temp - si) / r;
  } else {
    if (r * si < 1) {
      /* cone intersects earth */
      if (si == 0) {
	/* trivial case - theta = 0 */
	*radius = 0.0;
	*height = r - 1.0;
      } else {
	*radius = sin(asin(r * si) - theta);
	*height = (*radius) * c / si;
      }
    } else {
      /* cone does not intersect earth */
      *radius = r * si * c;
      *height = r * c * c;
    }
  }
}

/*
 * cone_write_geom
 *
 * Write cone geom for a satellite
 */
static void
cone_write_geom(const Satellite s, const Constellation * pconstellation)
{
  static char line[1000];
  unsigned int i;
  double t[16];

  for (i = 0; i < 16; i++)
    t[i] = 0.0;
  /*
     compute scaling of cone -
     a rigid body rotation plus a translation
     will be applied to the cone to place it so
     that (0,0,0) will be mapped to the position of the satellite
     and (0,0,-1) will point to the origin
   */
  cone_size(t, t + 10, s, pconstellation);
  t[5] = t[0];
  t[15] = 1.0;

  /* start of gv cone description */
  gv_begin();
  sprintf(line,
	  "(read geometry {define cone_%d { INST transform {%g %g %g %g %g %g %g %g %g %g %g %g %g %g %g %g} geom :cone_h} } )\n",
	  s->id, t[0], t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8], t[9],
	  t[10], t[11], t[12], t[13], t[14], t[15]);
  gv_send(line);
  gv_end();
}


/*
 * cone_display
 *
 * Takes a satellite and displays its cone (if turned on).
 */
void
cone_display(Satellite s, const Constellation * pconstellation)
{
  if (!cones_on_flag)
    return;
  cones_geom_needs_updating = TRUE;

  cone_write_geom(s, pconstellation);
}




/*
 * cone_delete
 *
 * Takes a satellite and deletes its cone
 * ONLY if cones_on_flag == TRUE
 */
void
cone_delete(Satellite s)
{
  static char name[] = "cone_XXXXXXXXXX";

  if (!cones_on_flag)
    return;
  cones_geom_needs_updating = TRUE;

  gv_begin();
  sprintf(name + 5, "%d", s->id);
  gv_delete_handle(name);
  gv_end();
}



void
write_cones_geom(const Constellation * pconstellation)
{
  Satellite s;
  Satellite_list sl = pconstellation->satellites;
  char buf[256];

  if (cones_on_flag && cones_geom_needs_updating) {
    /* write out new list of all satellites */
    gv_send("(geometry Cones {LIST\n");
    while (sl) {
      s = sl->s;
      if (sl->s->can_display_cone) {
	sprintf(buf, "{ INST transform:trans_%d geom:cone_%d}\n", s->id,
		s->id);
	gv_send(buf);
      }
      sl = sl->next;
    }
    gv_send("})\n");

    cones_geom_exists = TRUE;
    cones_geom_needs_updating = FALSE;
  }
}
