/*
 *****************************************************
 *
 *  SaVi by Robert Thurman (thurman@geom.umn.edu) and
 *          Patrick Worfolk (worfolk@alum.mit.edu).
 *
 *  Copyright (c) 1997 by The Geometry Center.
 *  This file is part of SaVi.  SaVi is free software;
 *  you can redistribute it and/or modify it only under
 *  the terms given in the file COPYRIGHT which you should
 *  have received along with this file.  SaVi may be
 *  obtained by emailing savi@geom.umn.edu.
 *
 *****************************************************
 *
 * earth.c
 *
 * Routines for creating fancy earth model
 *
 * $Id: earth.c,v 1.3 2002/12/27 21:01:37 lloyd Exp $
 */
#include <stdio.h>
#include <math.h>

#include "gv_utils.h"
#include "utils.h"
#include "constants.h"
#include "globals.h"
#include "savi.h"

static unsigned int earth_on_flag = FALSE;
static unsigned int earth_geom_exists = FALSE;
static unsigned int texture_flag = FALSE;
static unsigned int use_fancy_earth = FALSE;
static unsigned int earth_texture_loaded = FALSE;

/*
 * earth_create_geom
 */
void
earth_create_geom()
{
  if (earth_geom_exists)
    return;

  if (geomview_module) {
    /* create a simple blue sphere for the earth */
    gv_begin();
    gv_create_geom("Central_Body", "central_t", "unit_sphere_h");
    gv_end();
  }
  earth_geom_exists = TRUE;
}

/*
 * earth_delete_geom
 */
void
earth_delete_geom()
{
  if (!earth_geom_exists)
    return;

  if (geomview_module) {
    gv_begin();
    gv_delete_geom("Central_Body");
    gv_end();
  }
  earth_geom_exists = FALSE;
}

/*
 * earth_on_cmd
 *
 * Displays the earth sphere, dull or detailed.
 */
char *
earth_on_cmd(int argc, char *argv[])
{
  if (earth_on_flag)
    return EMPTY_str;

  earth_on_flag = TRUE;

  if (use_fancy_earth) {
    if (geomview_module) {
      gv_begin();
      if (texture_flag) {
	if (!earth_texture_loaded) {
	  earth_texture_loaded = gv_sendfile("oogl/earth_h.oogl");
	}
        if (earth_texture_loaded) {
	  gv_create_geom("Central_Body", "central_t", "earth_h");
	  earth_geom_exists = TRUE;
	} else {
	  error("could not load earth texture file.");
	}
      } else {
	gv_create_geom("Central_Body", "central_t", "earth_vect_sphere_h");
	earth_geom_exists = TRUE;
      }
      gv_end();
    }
  } else {
    /* just draw sphere */
    earth_create_geom();
  }

  return EMPTY_str;
}

/*
 * earth_off_cmd
 *
 * Destroy the earth sphere
 */
char *
earth_off_cmd(int argc, char *argv[])
{
  if (!earth_on_flag)
    return EMPTY_str;

  earth_on_flag = FALSE;

  if (use_fancy_earth) {
      /* we nest in gv_begin/end to hide temporary cosmetic changes -
       * this is the same technique used when loading tcl scripts
       * to prevent incremental per-new-satellite drawing updates.
       */
      if (geomview_module) gv_begin();
      fancy_off_cmd(argc, argv);
      fancy_on_cmd(argc, argv);
      if (geomview_module) gv_end();
  } else {
     /* just destroy sphere */
     earth_delete_geom();
  }

  return EMPTY_str;
}

/*
 * fancy_on_cmd
 *
 * Creates a fancy model of the earth by using vector
 * or loading texture-map file.
 */
char *
fancy_on_cmd(int argc, char *argv[])
{
  if (use_fancy_earth)
    return EMPTY_str;

  use_fancy_earth = TRUE;

  if (geomview_module) {
    gv_begin();
    if (earth_on_flag) {
      if (texture_flag) {
	if (!earth_texture_loaded) {
	  earth_texture_loaded = gv_sendfile("oogl/earth_h.oogl");
	}
        if (earth_texture_loaded) {
	  gv_create_geom("Central_Body", "central_t", "earth_h");
	  earth_geom_exists = TRUE;
	} else {
	  error("could not load earth texture file.");
	}
      } else {
	gv_create_geom("Central_Body", "central_t", "earth_vect_sphere_h");
	earth_geom_exists = TRUE;
      }
    } else {
      /*
       * draw only outline continents - ideal for spherical and
       * hyperbolic projections.
       */
      gv_create_geom("Central_Body", "central_t", "earth_vect_clear_h");
      earth_geom_exists = TRUE;
    }
    gv_end();
  }

  return EMPTY_str;
}


/*
 * fancy_off_cmd
 *
 * Returns to a simple sphere to model the central body
 */
char *
fancy_off_cmd(int argc, char *argv[])
{
  char *status = EMPTY_str;

  if (!use_fancy_earth)
    return EMPTY_str;

  use_fancy_earth = FALSE;

  if (earth_on_flag) {
    if (geomview_module) gv_begin();
    status = earth_off_cmd(argc, argv);
    status = earth_on_cmd(argc, argv);
    if (geomview_module) gv_end();
  } else {
    earth_delete_geom();
  }

  return status;
}

/*
 * texture_on
 *
 * Use texture mapping for fancy earth model
 */
char *
texture_on_cmd(int argc, char *argv[])
{
  char *status = EMPTY_str;

  texture_flag = TRUE;

  if (earth_on_flag) {
    if (geomview_module) gv_begin();
    status = earth_off_cmd(argc, argv);
    status = earth_on_cmd(argc, argv);
    if (geomview_module) gv_end();
  }

  return status;
}

/*
 * texture_off_cmd
 *
 * Use vect model for fancy earth model
 */
char *
texture_off_cmd(int argc, char *argv[])
{
  char *status = EMPTY_str;

  texture_flag = FALSE;

  if (earth_on_flag) {
    if (geomview_module) gv_begin();
    status = earth_off_cmd(argc, argv);
    status = earth_on_cmd(argc, argv);
    if (geomview_module) gv_end();
  }

  return status;
}


/*
 * earth_place
 *
 * rotates the earth to correct position for time t
 */
void
earth_place(double t, const CentralBody * pcb)
{
  double m[4][4], theta, ct, st;

  /*
   * Don't waste time rotating nothing.
   */
  if (!earth_geom_exists)
    return;

  /*
   * Don't waste time rotating a featureless blue sphere.
   */
  if (!use_fancy_earth)
    return;

  identity(m);

  theta = t * pcb->rotation_rate;

  ct = cos(theta);
  st = sin(theta);

  m[0][0] = ct;
  m[0][1] = st;
  m[1][0] = -st;
  m[1][1] = ct;

  gv_transform("central_t", m);

}
